/*
**==============================================================================
**
** Copyright (c) Microsoft Corporation. All rights reserved. See file LICENSE
** for license information.
**
**==============================================================================
*/

#if !defined(_SHELL_API_H_)
#define _SHELL_API_H_

#include <pal/palcommon.h>
#include <MI.h>

/* Error codes needed for compatibility with Windows WinRM */
#define ERROR_WSMAN_SERVICE_STREAM_DISCONNECTED 0x803381DE
#define ERROR_WSMAN_REDIRECT_REQUESTED 0x80338199
#define ERROR_INSUFFICIENT_BUFFER 122

/* NOTE: All strings need to be UTF-16. */
typedef void * HANDLE;
//
// -----------------------------------------------------------------------------
//  Structure for a data object.  These structures are used as output and input
//  to operations.  Eg, output from a WSManReceiveShellOutput and input to a
//      WSManSendShellInput.
// -----------------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------------
// data as text
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_DATA_TEXT
{
    MI_Uint32 bufferLength;
    _In_reads_(bufferLength) const MI_Char16 * buffer;
}WSMAN_DATA_TEXT;

//
// -----------------------------------------------------------------------------
// binary data used by API like for example WSManSendShellInput and
//  WSManReceiveShellOutput for Shell API, the data can be stream text
//  (ANSI/UNICODE), binary content or objects or partial or full XML
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_DATA_BINARY
{
    MI_Uint32 dataLength;
    _In_reads_(dataLength) MI_Uint8 *data;
}WSMAN_DATA_BINARY;

//
// -----------------------------------------------------------------------------
// type of the data used by the WSMAN_DATA structure
// -----------------------------------------------------------------------------
//

enum WSManDataType
{
    WSMAN_DATA_NONE                = 0,
    WSMAN_DATA_TYPE_TEXT           = 1,
    WSMAN_DATA_TYPE_BINARY         = 2,
    WSMAN_DATA_TYPE_DWORD          = 4
};
typedef enum WSManDataType WSManDataType;

//
// -----------------------------------------------------------------------------
// Structure used to pass data as XML text, binary data or MI_Uint32
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_DATA
{
    WSManDataType type;
    union
    {
        WSMAN_DATA_TEXT text;
        WSMAN_DATA_BINARY binaryData;
        MI_Uint32 number;
    };
} WSMAN_DATA;

//
// -----------------------------------------------------------------------------
//  Error structure containing fault code (a Windows error code that maps to
//  a SOAP fault) and extended error information.
//  The extended error information is the soap fault description from the protocol layer
//  including the machine where the fault was reported as well as plug-in
//  specific error text.
//
//  This error structure is used by callbacks to return detailed error information.
//
//  The error structure can contain also a transport error.
// -----------------------------------------------------------------------------
//

typedef struct _WSMAN_ERROR
{
    MI_Uint32 code;
    // extended error information
    _In_opt_ const MI_Char16* errorDetail; // extended error description from the fault;
                                 // it can be NULL, for example in out of memory conditions
    _In_opt_ const MI_Char16* language;    // language for error description (RFC 3066 language code); it can be NULL
    _In_opt_ const MI_Char16* machineName; // machine id; it can be NULL
    _In_opt_ const MI_Char16* pluginName;  // Plug-in name for errors generated by plug-ins. Otherwise NULL.
} WSMAN_ERROR;

//
// -----------------------------------------------------------------------------
// Username and password on the remote machine
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_USERNAME_PASSWORD_CREDS
{
    _In_ const MI_Char16* username;
    _In_ const MI_Char16* password;
} WSMAN_USERNAME_PASSWORD_CREDS;

//
// -----------------------------------------------------------------------------
// flags used by the authenticationMechanism member of WSMAN_AUTHENTICATION_CREDENTIALS structure
// -----------------------------------------------------------------------------
//
enum WSManAuthenticationFlags
{
    WSMAN_FLAG_DEFAULT_AUTHENTICATION    = 0x0,     //Use the default authentication
    WSMAN_FLAG_NO_AUTHENTICATION         = 0x1,     //Use no authentication for a remote operation
    WSMAN_FLAG_AUTH_DIGEST               = 0x2,     //Use digest authentication for a remote operation
    WSMAN_FLAG_AUTH_NEGOTIATE            = 0x4,     //Use negotiate authentication for a remote operation (may use kerberos or ntlm)
    WSMAN_FLAG_AUTH_BASIC                = 0x8,     //Use basic authentication for a remote operation
    WSMAN_FLAG_AUTH_KERBEROS             = 0x10,    //Use kerberos authentication for a remote operation
#if (WINVER >= 0x600)
    WSMAN_FLAG_AUTH_CREDSSP              = 0x80,    //Use CredSSP authentication for a remote operation
#endif
    WSMAN_FLAG_AUTH_CLIENT_CERTIFICATE   = 0x20     //Use client certificate authentication for a remote operation
};

//
// -----------------------------------------------------------------------------
// Structure for passing the credentials and the authentication mechanism
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_AUTHENTICATION_CREDENTIALS
{
    MI_Uint32 authenticationMechanism; // can be 0 (the user did not specify an
                                   // authentication mechanism, WSMan client
                                   // will choose between Kerberos and Negotiate only);
                                   // if it is not 0, it must be one of the
                                   // values from WSManAuthenticationFlags
                                   // enumeration
    union
    {
        WSMAN_USERNAME_PASSWORD_CREDS userAccount;
        const MI_Char16* certificateThumbprint;
    };
}WSMAN_AUTHENTICATION_CREDENTIALS;

//
// -----------------------------------------------------------------------
// Options
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_OPTION
{
    const MI_Char16 * name;
    const MI_Char16 * value;
    BOOL mustComply;
} WSMAN_OPTION;

typedef struct _WSMAN_OPTION_SET
{
    MI_Uint32 optionsCount;
    _In_reads_opt_(optionsCount) WSMAN_OPTION *options;
    BOOL optionsMustUnderstand;
} WSMAN_OPTION_SET;

typedef struct _WSMAN_OPTION_SETEX
{
    MI_Uint32 optionsCount;
    _In_reads_opt_(optionsCount) WSMAN_OPTION *options;
    BOOL optionsMustUnderstand;
    _In_reads_opt_(optionsCount) const MI_Char16* *optionTypes;
} WSMAN_OPTION_SETEX;

//
// -----------------------------------------------------------------------------
//  Structures containing information for addressing the endpoint.
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_KEY
{
    const MI_Char16 * key;
    const MI_Char16 * value;
} WSMAN_KEY;

typedef struct _WSMAN_SELECTOR_SET
{
    MI_Uint32 numberKeys; // Number of keys (selectors)
    _In_reads_opt_(numberKeys) WSMAN_KEY *keys;  // Array of key names and values
} WSMAN_SELECTOR_SET;

//
// -----------------------------------------------------------------------------
// Structure reserved for future use (Fragment Level WS-Transfer)
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_FRAGMENT
{
    _In_ const MI_Char16 * path;               // fragment path - WS-Transfer
    _In_opt_ const MI_Char16 * dialect;        // dialect for Fragment path

} WSMAN_FRAGMENT;

//
// -----------------------------------------------------------------------------
// Structure reserved for future use (Filter Enumeration/Eventing)
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_FILTER
{
    _In_ const MI_Char16 * filter;              // filter enumeration/subscription - allows ad-hoc queries using quey languages like SQL
    _In_opt_ const MI_Char16 * dialect;         // dialect for filter predicate

} WSMAN_FILTER;

#define WSMAN_OPERATION_INFOV1  0x00000000
#define WSMAN_OPERATION_INFOV2  0xaabbccdd

typedef struct _WSMAN_OPERATION_INFO
{
    _In_opt_ WSMAN_FRAGMENT fragment;              // optional element to support Fragment transfer or
    _In_opt_ WSMAN_FILTER filter;                  // optional Filter WS-Enumerate/WS-Eventing
    _In_opt_ WSMAN_SELECTOR_SET selectorSet;
    _In_opt_ WSMAN_OPTION_SET optionSet;
    _In_opt_ void *reserved;
             MI_Uint32 version;
} WSMAN_OPERATION_INFO;

typedef struct _WSMAN_OPERATION_INFOEX
{
    _In_opt_ WSMAN_FRAGMENT fragment;              // optional element to support Fragment transfer or
    _In_opt_ WSMAN_FILTER filter;                  // optional Filter WS-Enumerate/WS-Eventing
    _In_opt_ WSMAN_SELECTOR_SET selectorSet;
    _In_opt_ WSMAN_OPTION_SETEX optionSet;
             MI_Uint32 version;
    _In_opt_ const MI_Char16* uiLocale;
    _In_opt_ const MI_Char16* dataLocale;
} WSMAN_OPERATION_INFOEX;
//
// -----------------------------------------------------------------------------
//  Client Initialization/Deinitialization functions
// -----------------------------------------------------------------------------
//

typedef struct WSMAN_API *WSMAN_API_HANDLE;

//
// -----------------------------------------------------------------------------
// This API is used to initialize the WinRM client;
// It can be used by different clients on the same process, ie svchost.exe.
// Returns a nonzero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManInitialize(
    MI_Uint32 flags,
    _Out_ WSMAN_API_HANDLE *apiHandle
    );

//
// -----------------------------------------------------------------------------
// This API deinitializes the Winrm client stack; all operations will
//  finish before this API will return; this is a sync call;
//  it is highly recommended that all operations are explictly cancelled
//  and all sessions are closed before calling this API
// Returns non zero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManDeinitialize(
    _Inout_opt_ WSMAN_API_HANDLE apiHandle,
    MI_Uint32 flags
    );

//
// -----------------------------------------------------------------------------
// This API returns a NULL terminated unicode string containing the message
//  of an error number and an optional language identifier. The optional
//  parameter languageCode specifies the UI language as RFC 3066 language code
//  that should be used to localize the message (if not specified, the thread's
//  UI language will be used). If the function cannot find a message for that
//  language, it returns ERROR_RESOURCE_LANG_NOT_FOUND. The function copies
//  the formatted message text to an output buffer.
//
// Returns non zero error code upon failure. If the output buffer is not
//  big enough or NULL, the function returns ERROR_INSUFFICIENT_BUFFER and
//  the messageLengthUsed parameter contains the requested size of the buffer.
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManGetErrorMessage(
    _In_ WSMAN_API_HANDLE apiHandle,
    MI_Uint32 flags,            // reserved for future use; must be 0
    _In_opt_ const MI_Char16* languageCode,    // the RFC 3066 language code; it can be NULL
    MI_Uint32 errorCode,        // error code for the requested error message
    MI_Uint32 messageLength,    // message length, including NULL terminator
    _Out_writes_to_opt_(messageLength, *messageLengthUsed) MI_Char16* message,
    _Out_ MI_Uint32* messageLengthUsed  // effective message length, including NULL terminator
    );


//
// -----------------------------------------------------------------------------
// Unsigned long integer value that contains the proxy access type
// By default, wsman uses WSMAN_OPTION_PROXY_WINHTTP_PROXY_CONFIG -
//  the proxy settings configured for WinHTTP. The WinHTTP proxy settings
//  can be set with Proxycfg.exe or by using netsh command.
// When WSMAN_OPTION_PROXY_IE_PROXY_CONFIG is specified, the current user's
//  Internet Explorer proxy settings for the current active network connection.
//  This option requires the user profile to be loaded, so the option can
//  be directly used when called within a process that is running under
//  an interactive user account identity; if the client application is running
//  under a user context different then the interactive user, the client
//  application has to explicitly load the user profile prior to using this option.
// IMPORTANT: proxy configuration is supported for HTTPS only; for HTTP, the direct
//      connection to the server is used
// -----------------------------------------------------------------------------
//
enum WSManProxyAccessType
{
    WSMAN_OPTION_PROXY_IE_PROXY_CONFIG          = 1,    // use the Internet Explorer proxy configuration for the current user
    WSMAN_OPTION_PROXY_WINHTTP_PROXY_CONFIG     = 2,    // default: proxy settings configured for WinHTTP, using the ProxyCfg.exe utility
    WSMAN_OPTION_PROXY_AUTO_DETECT              = 4,    // Force autodetection of proxy
    WSMAN_OPTION_PROXY_NO_PROXY_SERVER          = 8,    // do not use a proxy server - resolves all host names locally
};

//
// -----------------------------------------------------------------------------
// Structure used to set the proxy information per session
// -----------------------------------------------------------------------------
//

typedef struct _WSMAN_PROXY_INFO
{
    MI_Uint32 accessType;
    _In_opt_ WSMAN_AUTHENTICATION_CREDENTIALS authenticationCredentials; // credentials and authentication scheme used for proxy
} WSMAN_PROXY_INFO;

//
// -----------------------------------------------------------------------------
//  Client Session
// -----------------------------------------------------------------------------
//
typedef struct WSMAN_SESSION *WSMAN_SESSION_HANDLE;

//
// -----------------------------------------------------------------------------
// Creates a session which can be used to perform subsequent operations
// Returns a non zero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManCreateSession(
    _In_ WSMAN_API_HANDLE apiHandle,
    _In_opt_ const MI_Char16* connection,             // if NULL, then connection will default to localhost
    MI_Uint32 flags,
    _In_opt_ WSMAN_AUTHENTICATION_CREDENTIALS *serverAuthenticationCredentials,
    _In_opt_ WSMAN_PROXY_INFO *proxyInfo,
    _Out_ WSMAN_SESSION_HANDLE *session
    );

//
// -----------------------------------------------------------------------------
// Frees memory of session and closes all related operations before returning; this is sync call
// it is recommended that all pending operations are either completed or cancelled before calling this API
// Returns a non zero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManCloseSession(
    _Inout_opt_ WSMAN_SESSION_HANDLE session,
    MI_Uint32 flags
    );

//
// default operation timeout for network operations - 1 min = 60000ms
//
#define WSMAN_DEFAULT_TIMEOUT_MS 60000

//
// -----------------------------------------------------------------------------
// Extended options set for the session
// -----------------------------------------------------------------------------
//
enum WSManSessionOption
{
    //
    //Timeouts
    //

    WSMAN_OPTION_DEFAULT_OPERATION_TIMEOUTMS = 1,// MI_Uint32 - default timeout in ms that applies to all operations on the client side

    WSMAN_OPTION_MAX_RETRY_TIME             = 11, // MI_Uint32 (read only) - maximum time for Robust connection retries
    WSMAN_OPTION_TIMEOUTMS_CREATE_SHELL      = 12,// MI_Uint32 - timeout in ms for WSManCreateShell operations
    WSMAN_OPTION_TIMEOUTMS_RUN_SHELL_COMMAND =13,// MI_Uint32 - timeout in ms for WSManRunShellCommand operations
    WSMAN_OPTION_TIMEOUTMS_RECEIVE_SHELL_OUTPUT =14,// MI_Uint32 - timeout in ms for WSManReceiveShellOutput operations
    WSMAN_OPTION_TIMEOUTMS_SEND_SHELL_INPUT  = 15,// MI_Uint32 - timeout in ms for WSManSendShellInput operations
    WSMAN_OPTION_TIMEOUTMS_SIGNAL_SHELL      = 16,// MI_Uint32 - timeout in ms for WSManSignalShell and WSManCloseCommand operations
    WSMAN_OPTION_TIMEOUTMS_CLOSE_SHELL       = 17,// MI_Uint32 - timeout in ms for WSManCloseShell operations

    //
    // connection options
    //

    WSMAN_OPTION_SKIP_CA_CHECK              = 18,// MI_Uint32  - 1 to not validate the CA on the server certificate; 0 - default
    WSMAN_OPTION_SKIP_CN_CHECK              = 19,// MI_Uint32  - 1 to not validate the CN on the server certificate; 0 - default
    WSMAN_OPTION_UNENCRYPTED_MESSAGES       = 20,// MI_Uint32  - 1 to not encrypt the messages; 0 - default
    WSMAN_OPTION_UTF16                      = 21,// MI_Uint32  - 1 Send all network packets for remote operatons in UTF16; 0 - default is UTF8
    WSMAN_OPTION_ENABLE_SPN_SERVER_PORT     = 22,// MI_Uint32  - 1 When using negotiate, include port number in the connection SPN; 0 - default
                                                 // Used when not talking to the main OS on a machine but, for instance, a BMC
    WSMAN_OPTION_MACHINE_ID                 = 23,// MI_Uint32  - 1 Identify this machine to the server by including the MachineID header; 0 - default

    //
    // other options
    //
    WSMAN_OPTION_LOCALE                     = 25,// string - RFC 3066 language code
    WSMAN_OPTION_UI_LANGUAGE                = 26,// string - RFC 3066 language code
    WSMAN_OPTION_MAX_ENVELOPE_SIZE_KB       = 28,// MI_Uint32 - max SOAP envelope size (kb) - default 150kb from winrm config
                                                // (see 'winrm help config' for more details); the client SOAP packet size cannot surpass
                                                //  this value; this value will be also sent to the server in the SOAP request as a
                                                //  MaxEnvelopeSize header; the server will use min(MaxEnvelopeSizeKb from server configuration,
                                                //  MaxEnvelopeSize value from SOAP).
    WSMAN_OPTION_SHELL_MAX_DATA_SIZE_PER_MESSAGE_KB = 29,// MI_Uint32 (read only) - max data size (kb) provided by the client, guaranteed by
                                                     //  the winrm client implementation to fit into one SOAP packet; this is an
                                                     // approximate value calculated based on the WSMAN_OPTION_MAX_ENVELOPE_SIZE_KB (default 500kb),
                                                     // the maximum possible size of the SOAP headers and the overhead of the base64
                                                     // encoding which is specific to WSManSendShellInput API; this option can be used
                                                     // with WSManGetSessionOptionAsDword API; it cannot be used with WSManSetSessionOption API.
    WSMAN_OPTION_REDIRECT_LOCATION          = 30,// string - read-only, cannot set
    WSMAN_OPTION_SKIP_REVOCATION_CHECK  = 31,// MI_Uint32  - 1 to not validate the revocation status on the server certificate; 0 - default
    WSMAN_OPTION_ALLOW_NEGOTIATE_IMPLICIT_CREDENTIALS  = 32,// MI_Uint32  - 1 to allow default credentials for Negotiate (this is for SSL only); 0 - default
    WSMAN_OPTION_USE_SSL                    = 33,    // MI_Uint32 - When using just a machine name in the connection string use an SSL connection. 0 means HTTP, 1 means HTTPS.  Default is 0.
    WSMAN_OPTION_USE_INTERACTIVE_TOKEN  = 34        // MI_Uint32 - When creating connection on local machine, use interactive token feature. 1 - default
};
typedef enum WSManSessionOption WSManSessionOption;

//
// -----------------------------------------------------------------------------
// WSManSetSessionOption API - set session options
//  Returns a non zero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManSetSessionOption(
    _In_ WSMAN_SESSION_HANDLE session,
    WSManSessionOption option,
    _In_ WSMAN_DATA *data
    );

//
// -----------------------------------------------------------------------------
// WSManGetSessionOptionAsDword API - get a session option
//  Returns a non zero error code upon failure
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManGetSessionOptionAsDword(
    _In_ WSMAN_SESSION_HANDLE session,
    WSManSessionOption option,
    _Inout_ MI_Uint32 *value
    );

//
// -----------------------------------------------------------------------------
// WSManGetSessionOptionAsString API - get a session option
//  Returns a non zero error code upon failure
// -----------------------------------------------------------------------------
//
_Success_(return == NO_ERROR)
MI_Uint32 WINAPI WSManGetSessionOptionAsString(
    _In_ WSMAN_SESSION_HANDLE session,
    WSManSessionOption option,
    MI_Uint32 stringLength,
    _Out_writes_to_opt_(stringLength, *stringLengthUsed) MI_Char16* string,
    _Out_ MI_Uint32* stringLengthUsed
    );

//
// -----------------------------------------------------------------------------
// Handle returned by WS-Transfer and WS-Enumerate operations
// -----------------------------------------------------------------------------
//
typedef struct WSMAN_OPERATION *WSMAN_OPERATION_HANDLE;

//
// -----------------------------------------------------------------------------
// flags used by WSMAN_SHELL_COMPLETION_FUNCTION callback function
// -----------------------------------------------------------------------------
//
enum WSManCallbackFlags
{
    //
    // Flag that marks the end of any single step of multistep operation
    //
    WSMAN_FLAG_CALLBACK_END_OF_OPERATION     = 0x1,

    //
    // WSMAN_SHELL_COMPLETION_FUNCTION API specific flags
    //  end of a particular stream; it is used for optimization purposes if the shell
    //  knows that no more output will occur for this stream; in some conditions this
    //  cannot be determined.
    //
    WSMAN_FLAG_CALLBACK_END_OF_STREAM        = 0x8,

    // Flag that if present on CreateShell callback indicates that it supports disconnect
    WSMAN_FLAG_CALLBACK_SHELL_SUPPORTS_DISCONNECT = 0x20,

    // Flag that indicates that the shell got disconnected due to netowrk failure
    WSMAN_FLAG_CALLBACK_SHELL_AUTODISCONNECTED     = 0x40,

    // Flag indicates that the client shell detected a network failure
    WSMAN_FLAG_CALLBACK_NETWORK_FAILURE_DETECTED = 0x100,

    // Flag indicates that client shell is retrying to establish network connection with the server
    WSMAN_FLAG_CALLBACK_RETRYING_AFTER_NETWORK_FAILURE = 0x200,

    // Flag indicates that client shell successfully reconnected with the server after attempting to reconnect to the server
    WSMAN_FLAG_CALLBACK_RECONNECTED_AFTER_NETWORK_FAILURE = 0x400,

    // Flag indicates that the client shell attempts to reconnect to the server failed and hence it is AutoDisconnecting
    WSMAN_FLAG_CALLBACK_SHELL_AUTODISCONNECTING      = 0x800,

    // Flag indicates that the client shell got into broken state in the middle of retry notification sequence due to some internal error at wsman layer
    WSMAN_FLAG_CALLBACK_RETRY_ABORTED_DUE_TO_INTERNAL_ERROR = 0x1000,

    // Flag that indicates for a receive operation that a delay stream request has been processed
    WSMAN_FLAG_CALLBACK_RECEIVE_DELAY_STREAM_REQUEST_PROCESSED = 0X2000
};

//
// -----------------------------------------------------------------------------
// Closes an asynchronous operation; if the callback associated with the operation
//  is pending and have not completed when WSManCloseOperation is called, then
//  the function marks the operation for deletion and returns; If the callback was not called,
//  the operation is cancelled and the operation callback is called with
//  WSMAN_ERROR_OPERATION_ABORTED error; the operation handle is freed in all cases
//  after the callback returns.
// -----------------------------------------------------------------------------
//
MI_Uint32 WINAPI WSManCloseOperation(
    _Inout_opt_ WSMAN_OPERATION_HANDLE operationHandle,
    MI_Uint32 flags
    );

//
//------------------------------------------------------------------------
// Shell Client API
//  The API set includes the ability to create a shell, execute one or
//  more commands, possibly pipe input stream to the command, collect the
//  output stream from the command, terminate the command and finally
//  close the shell. The client must retrieve the output initiating
//  a WSManReceiveShellOutput once; the underlying implementation will
//  handle pulling the data from the command, until the command is done.
//  For each received stream, the completion function will be called with
//  the status of the command.
//
//  WSManCloseCommand or WSManCloseShell can be called any time to close the
//    commmand or a shell operation. The callback will be called with
//    WSMAN_FLAG_CALLBACK_END_OF_OPERATION flag to mark the end of the operation.
//    See WSManCloseCommand/WSManCloseShell API description for more info.
// -----------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------
// Shell handle returned by WSManCreateShell API
// -----------------------------------------------------------------------
//
typedef struct WSMAN_SHELL *WSMAN_SHELL_HANDLE;

//
// -----------------------------------------------------------------------
// Command handle returned by WSManRunShellCommand API
// -----------------------------------------------------------------------
//
typedef struct WSMAN_COMMAND *WSMAN_COMMAND_HANDLE;

//
// -----------------------------------------------------------------------
// predefined stream ids
// -----------------------------------------------------------------------
//
#define WSMAN_STREAM_ID_STDIN PAL_T("stdin")
#define WSMAN_STREAM_ID_STDOUT PAL_T("stdout")
#define WSMAN_STREAM_ID_STDERR PAL_T("stderr")

//
// -----------------------------------------------------------------------
// stream selector - array of stream ids; ie "stdin" or "stdout" or "stderr"
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_STREAM_ID_SET
{
    MI_Uint32 streamIDsCount;
    _In_reads_opt_(streamIDsCount) const MI_Char16* *streamIDs;

} WSMAN_STREAM_ID_SET;

//
// -----------------------------------------------------------------------
// Environment variable
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_ENVIRONMENT_VARIABLE
{
    const MI_Char16* name;
    const MI_Char16* value;

} WSMAN_ENVIRONMENT_VARIABLE;

//
// -----------------------------------------------------------------------
// environment variable set - array of environment variables
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_ENVIRONMENT_VARIABLE_SET
{
    MI_Uint32 varsCount;
    _In_reads_opt_(varsCount) WSMAN_ENVIRONMENT_VARIABLE *vars;

} WSMAN_ENVIRONMENT_VARIABLE_SET;

//
// -----------------------------------------------------------------------
// Shell initialization parameters
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_SHELL_STARTUP_INFO_V11
{
    WSMAN_STREAM_ID_SET *inputStreamSet;
    WSMAN_STREAM_ID_SET *outputStreamSet;
    MI_Uint32 idleTimeoutMs;
    const MI_Char16* workingDirectory;
    WSMAN_ENVIRONMENT_VARIABLE_SET *variableSet;

    const MI_Char16* name;
} WSMAN_SHELL_STARTUP_INFO_V11;


typedef WSMAN_SHELL_STARTUP_INFO_V11 WSMAN_SHELL_STARTUP_INFO;

typedef struct _WSMAN_SHELL_DISCONNECT_INFO
{
    MI_Uint32 idleTimeoutMs;
} WSMAN_SHELL_DISCONNECT_INFO;

//
// -----------------------------------------------------------------------
// Pre-defined URIs
// -----------------------------------------------------------------------
//
#define WSMAN_SHELL_NS PAL_T("http://schemas.microsoft.com/wbem/wsman/1/windows/shell")
#define WSMAN_SHELL_NS_LEN (sizeof(WSMAN_SHELL_NS)/sizeof(MI_Char)-1)

//
// -----------------------------------------------------------------------
// URI to create a Windows command shell; used with WSManCreateShell API
// -----------------------------------------------------------------------
//
#define WSMAN_CMDSHELL_URI WSMAN_SHELL_NS "/cmd"


//
// -----------------------------------------------------------------------
// Windows shell options
// -----------------------------------------------------------------------
//
enum WSManShellFlag
{
    //Turn off compression for Send/Receive operations.  By default compression is
    //turned on, but if communicating with a down-level box it may be necessary to
    //do this.  Other reasons for turning it off is due to the extra memory consumption
    //and CPU utilization that is used as a result of compression.
    WSMAN_FLAG_NO_COMPRESSION = 0x1,
    WSMAN_FLAG_DELETE_SERVER_SESSION = 0x2,
    //Enable the service to drop operation output when running disconnected
    WSMAN_FLAG_SERVER_BUFFERING_MODE_DROP = 0x4,
    //Enable the service to block operation progress when output buffers are full
    WSMAN_FLAG_SERVER_BUFFERING_MODE_BLOCK = 0x8,
    //Enable receive call to not immediately retrieve results. Only applicable for Receive calls on commands
    WSMAN_FLAG_RECEIVE_DELAY_OUTPUT_STREAM = 0X10
};
typedef enum WSManShellFlag WSManShellFlag;

//
// -----------------------------------------------------------------------
// Windows command shell specific options
// -----------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------
// Code page option name to be used with WSManCreateShell
//  API to remotely set the code page
// -----------------------------------------------------------------------
//
#define WSMAN_CMDSHELL_OPTION_CODEPAGE MI("WINRS_CODEPAGE")

//
// -----------------------------------------------------------------------
// Option name used with WSManCreateShell API to not load the
//  user profile on the remote server
// -----------------------------------------------------------------------
//
#define WSMAN_SHELL_OPTION_NOPROFILE PAL_T("WINRS_NOPROFILE")

//
// -----------------------------------------------------------------------
// Option name used with WSManRunShellCommand API to
//  indicate that the client side mode of standard input is Console;
//  default  implies Pipe.
// -----------------------------------------------------------------------
//
#define WSMAN_CMDSHELL_OPTION_CONSOLEMODE_STDIN PAL_T("WINRS_CONSOLEMODE_STDIN")

//
// -----------------------------------------------------------------------
// to be used with WSManRunShellCommand API to not use cmd.exe /c prefix when launching the command
// -----------------------------------------------------------------------
//
#define WSMAN_CMDSHELL_OPTION_SKIP_CMD_SHELL PAL_T("WINRS_SKIP_CMD_SHELL")

//
// -----------------------------------------------------------------------
// pre-defined command states
// -----------------------------------------------------------------------
//
#define WSMAN_COMMAND_STATE_DONE WSMAN_SHELL_NS PAL_T("/CommandState/Done")
#define WSMAN_COMMAND_STATE_PENDING WSMAN_SHELL_NS PAL_T("/CommandState/Pending")
#define WSMAN_COMMAND_STATE_RUNNING WSMAN_SHELL_NS PAL_T("/CommandState/Running")

//
// -----------------------------------------------------------------------
// Data structure returned after a Receive operation; each WSMAN_RECEIVE_DATA_RESULT
//  has a stream element with data that can be text (ANSI/UNICODE), binary content,
//  objects, or partial or full XML. In addition, the command state is reported
//  to the client and exitCode if the command is done.
//
// This structure is allocated and owned by the winrm stack and it is valid
//  in the completion function only
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_RECEIVE_DATA_RESULT
{
    const MI_Char16* streamId;          // string - any output stream name from the list passed to
                              // the server when creating the shell
    WSMAN_DATA streamData;    // always data as binary; can be stream text (ANSI/UNICODE), binary content or objects or partial or full XML

    // command state string values (shell specific) - the state of the command for which status is being reported
    const MI_Char16* commandState;      // if WSMAN_COMMAND_STATE_DONE then the command should be closed at this moment
    MI_Uint32 exitCode;

} WSMAN_RECEIVE_DATA_RESULT;


typedef struct _WSMAN_CONNECT_DATA
{
    WSMAN_DATA data;
} WSMAN_CONNECT_DATA;

typedef struct _WSMAN_CREATE_SHELL_DATA
{
    WSMAN_DATA data;
} WSMAN_CREATE_SHELL_DATA;

typedef union _WSMAN_RESPONSE_DATA
{
    WSMAN_RECEIVE_DATA_RESULT receiveData;
    WSMAN_CONNECT_DATA connectData;
    WSMAN_CREATE_SHELL_DATA createData;
} WSMAN_RESPONSE_DATA;

//
// -----------------------------------------------------------------------
// Completion function used by all Shell functions
// Returns error->code != 0 upon error; use error->errorDetail structure
//  for extended error informations; the callback is called for each
//  shell operation; after a WSManReceiveShellOutput operation is initiated,
//  the callback is called for each output stream element or if error;
//  the underlying implementation handles the polling of stream data
//  from the command or shell. If WSMAN_COMMAND_STATE_DONE state is received, no more
//  streams will be received from the command, so the command can be closed
//  using WSManCloseCommand(command).
//
// If error->code != 0, the result is guaranteed to be NULL
//
// The error and result objects are allocated and owned by the winrm
//  client stack; they are valid during the callback only; the user
//  has to synchronously copy the data in the callback
//
// This callback function will use the current access token, whether it is
//  a process or impersonation token.
// -----------------------------------------------------------------------
//


typedef void (*WSMAN_SHELL_COMPLETION_FUNCTION)(
    _In_opt_ PVOID operationContext,                     //user defined context
    MI_Uint32 flags,                                         // one or more flags from WSManCallbackFlags
    _In_ WSMAN_ERROR *error,                             // error allocated and owned by the winrm stack; valid in the callback only;
    _In_ WSMAN_SHELL_HANDLE shell,                       // shell handle associated with the user context; must be closed using WSManCloseShell
    _In_opt_ WSMAN_COMMAND_HANDLE command,               // command handle associated with the user context; must be closed using WSManCloseCommand
    _In_opt_ WSMAN_OPERATION_HANDLE operationHandle,     // valid only for Send/Receive/Signal operations; must be closed using WSManCloseOperation
    _In_opt_ WSMAN_RESPONSE_DATA *data                   // output data from command/shell; allocated internally and owned by the winrm stack
                                                         // valid only within this function. Currently used to send back receive stream data and
                                                         // open Xml content in rsp:ConnectResponse
    );
//
// -----------------------------------------------------------------------------
// Asynchronous structure to be passed to all Shell operations;
//  it contains an optional user context and the callback function
// -----------------------------------------------------------------------------
//
typedef struct _WSMAN_SHELL_ASYNC
{
    _In_opt_ PVOID operationContext;
    _In_ WSMAN_SHELL_COMPLETION_FUNCTION completionFunction;
} WSMAN_SHELL_ASYNC;

//
// -----------------------------------------------------------------------------
//  WSManRunShellCommand API - rsp:Command
// -----------------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------
// array of command arguments
// -----------------------------------------------------------------------
//
typedef struct _WSMAN_COMMAND_ARG_SET
{
    MI_Uint32 argsCount;
    _In_reads_opt_(argsCount) const MI_Char16* *args;

} WSMAN_COMMAND_ARG_SET;

//
// -----------------------------------------------------------------------------
//  WSManSignalShell API - rsp:Signal
// -----------------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------
// control codes for the command
// -----------------------------------------------------------------------
//
#define WSMAN_SIGNAL_SHELL_CODE_TERMINATE WSMAN_SHELL_NS PAL_T("/signal/terminate")
#define WSMAN_SIGNAL_SHELL_CODE_CTRL_C WSMAN_SHELL_NS PAL_T("/signal/ctrl_c")
#define WSMAN_SIGNAL_SHELL_CODE_CTRL_BREAK WSMAN_SHELL_NS PAL_T("/signal/ctrl_break")

void WINAPI WSManSignalShell(

    _In_ WSMAN_SHELL_HANDLE shell,
    _In_opt_ WSMAN_COMMAND_HANDLE command,         // if NULL, the Signal will be sent to the shell
    MI_Uint32 flags,
    _In_ const MI_Char16* code,             // signal code
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_OPERATION_HANDLE *signalOperation  // should be closed using WSManCloseOperation
);

//
// -----------------------------------------------------------------------------
//  WSManReceiveShellOutput API - rsp:Receive
// -----------------------------------------------------------------------------
//

void WINAPI WSManReceiveShellOutput(

    _Inout_ WSMAN_SHELL_HANDLE shell,
    _In_opt_ WSMAN_COMMAND_HANDLE command,
    MI_Uint32 flags,
    _In_opt_ WSMAN_STREAM_ID_SET *desiredStreamSet,  // request output from a particular stream or list of streams
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_OPERATION_HANDLE *receiveOperation // should be closed using WSManCloseOperation
);

//
// -----------------------------------------------------------------------------
//  WSManSendShellInput API - rsp:Send
// -----------------------------------------------------------------------------
//
void WINAPI WSManSendShellInput(

    _In_ WSMAN_SHELL_HANDLE shell,
    _In_opt_ WSMAN_COMMAND_HANDLE command,
    MI_Uint32 flags,
    _In_ const MI_Char16* streamId,               // input stream name
    _In_ WSMAN_DATA *streamData,        // data as binary - that can contain text (ANSI/UNICODE),
                                        // binary content or objects or partial or full XML
    BOOL endOfStream,
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_OPERATION_HANDLE *sendOperation // should be closed using WSManCloseOperation
);

//
// -----------------------------------------------------------------------------
// WSManCloseCommand API
// -----------------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------------
// Closes a command (signals the termination of a command); the WSManCloseCommand callback
//  is called with WSMAN_FLAG_CALLBACK_END_OF_OPERATION flag as result of this operation
// -----------------------------------------------------------------------------
//
void WINAPI WSManCloseCommand(
    _Inout_opt_ WSMAN_COMMAND_HANDLE commandHandle,
    MI_Uint32 flags,
    _In_ WSMAN_SHELL_ASYNC *async
    );

//
// -----------------------------------------------------------------------------
// WSManCloseShell API
// -----------------------------------------------------------------------------
//

//
// -----------------------------------------------------------------------------
// Closes a shell; the WSManCloseShell callback is called with
//  WSMAN_FLAG_CALLBACK_END_OF_OPERATION flag as result of this operation
// -----------------------------------------------------------------------------
//
void WINAPI WSManCloseShell(
    _Inout_opt_ WSMAN_SHELL_HANDLE shellHandle,
    MI_Uint32 flags,
    _In_ WSMAN_SHELL_ASYNC *async
    );


//
// -----------------------------------------------------------------------------
// WSManCreateShellEx API - wxf:Create with specific shell Id.
// -----------------------------------------------------------------------------
//
void WINAPI WSManCreateShellEx(
    _Inout_ WSMAN_SESSION_HANDLE session,
    MI_Uint32 flags,
    _In_ const MI_Char16* resourceUri,           // shell resource URI
    _In_ const MI_Char16* shellId,
    _In_opt_ WSMAN_SHELL_STARTUP_INFO *startupInfo,
    _In_opt_ WSMAN_OPTION_SET *options,
    _In_opt_ WSMAN_DATA *createXml,                     // open content for create shell
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_SHELL_HANDLE *shell // should be closed using WSManCloseShell
);

//
// -----------------------------------------------------------------------------
// WSManRunShellCommandEx API - rsp:Command with specific command Id.
// -----------------------------------------------------------------------------
//
void WINAPI WSManRunShellCommandEx(

    _Inout_ WSMAN_SHELL_HANDLE shell,
    MI_Uint32 flags,
    _In_ const MI_Char16* commandId,
    _In_ const MI_Char16* commandLine,
    _In_opt_ WSMAN_COMMAND_ARG_SET *args,
    _In_opt_ WSMAN_OPTION_SET *options,
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_COMMAND_HANDLE *command // should be closed using WSManCloseCommand
);

//
// -----------------------------------------------------------------------------
// WSManDisconnectShell API - rsp:Disconnect
// -----------------------------------------------------------------------------
//
void WINAPI WSManDisconnectShell(
    _Inout_ WSMAN_SHELL_HANDLE shell,
    MI_Uint32 flags,
    _In_ WSMAN_SHELL_DISCONNECT_INFO* disconnectInfo,
    _In_ WSMAN_SHELL_ASYNC *async
    );

//
// -----------------------------------------------------------------------------
// WSManReconnectShell API - rsp:Disconnect
// -----------------------------------------------------------------------------
//
void WINAPI WSManReconnectShell(
    _Inout_ WSMAN_SHELL_HANDLE shell,
    MI_Uint32 flags,
    _In_ WSMAN_SHELL_ASYNC *async
    );

//
// -----------------------------------------------------------------------------
// WSManReconnectShell API - rsp:Disconnect
// This is a temporary API for first week's delivery. This will be replaced
// by an API that does not take a callback. Instead the callback supplied to
// WSManRunShellCommand is used to report the completion of this operation
// -----------------------------------------------------------------------------
//
void WINAPI WSManReconnectShellCommand(
    _Inout_ WSMAN_COMMAND_HANDLE commandHandle,
    MI_Uint32 flags,
    _In_ WSMAN_SHELL_ASYNC *async
    );

//
// -----------------------------------------------------------------------------
// WSManConnectShell API - rsp:Connect
// -----------------------------------------------------------------------------
//
void WINAPI WSManConnectShell(
    _Inout_ WSMAN_SESSION_HANDLE session,
    MI_Uint32 flags,
    _In_ const MI_Char16* resourceUri,
    _In_ const MI_Char16* shellID,           // shell Identifier
    _In_opt_ WSMAN_OPTION_SET *options,
    _In_opt_ WSMAN_DATA *connectXml,                     // open content for connect shell
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_SHELL_HANDLE *shell // should be closed using WSManCloseShell
);

//
// -----------------------------------------------------------------------------
// WSManConnectShellCommand API - rsp:Connect
// -----------------------------------------------------------------------------
//
void WINAPI WSManConnectShellCommand(
    _Inout_ WSMAN_SHELL_HANDLE shell,
    MI_Uint32 flags,
    _In_ const MI_Char16* commandID,  //command Identifier
    _In_opt_ WSMAN_OPTION_SET *options,
    _In_opt_ WSMAN_DATA *connectXml,                // open content for connect command
    _In_ WSMAN_SHELL_ASYNC *async,
    _Out_ WSMAN_COMMAND_HANDLE *command // should be closed using WSManCloseCommand
);










typedef struct _WSMAN_CERTIFICATE_DETAILS
{
    const MI_Char16 * subject;         // Certificate subject in distinguished form
                            // Ex: "CN=xyz.com, OU=xyz management, O=Microsoft, L=Redmond, S=Washington, C=US"
    const MI_Char16 * issuerName;      // Certificate issuer in distinguished form
                            // Ex: "CN=Microsoft Secure Server Authority, DC=redmond, DC=corp, DC=microsoft, DC=com"
    const MI_Char16 * issuerThumbprint;// Thumbprint of Certificate issuer
    const MI_Char16 * subjectName;     // Certificate Subject Alternative Name (SAN) if available or subject Common Name (CN)
                            // Ex: "xyz.com"
} WSMAN_CERTIFICATE_DETAILS;

typedef struct _WSMAN_SENDER_DETAILS
{
    const MI_Char16 * senderName;                  // Username of the sender
    const MI_Char16 * authenticationMechanism;      //
    WSMAN_CERTIFICATE_DETAILS *certificateDetails; // valid only if the authentication is client certificates
    HANDLE clientToken;
    const MI_Char16 * httpURL;
} WSMAN_SENDER_DETAILS;

typedef struct _WSMAN_PLUGIN_REQUEST
{
    WSMAN_SENDER_DETAILS *senderDetails;
    const MI_Char16 * locale;
    const MI_Char16 * resourceUri;
    WSMAN_OPERATION_INFO *operationInfo;
    volatile BOOL shutdownNotification;
    HANDLE shutdownNotificationHandle;
    const MI_Char16 * dataLocale;
} WSMAN_PLUGIN_REQUEST;



/* ===================================================================
 *  * Following APIs are callbacks from the _PLUGIN_
 * APIs back into the infrastructure to report
 * results and contexts, or to query extra
 * information about the shell
 * ===================================================================
 */
MI_Uint32 MI_CALL WSManPluginReportContext(
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void * context
    );

#define WSMAN_FLAG_RECEIVE_RESULT_NO_MORE_DATA 1
#define WSMAN_FLAG_RECEIVE_FLUSH               2
#define WSMAN_FLAG_RECEIVE_RESULT_DATA_BOUNDARY 4

MI_Uint32 MI_CALL WSManPluginReceiveResult(
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_opt_ const MI_Char16 * stream,
    _In_opt_ WSMAN_DATA *streamResult,
    _In_opt_ const MI_Char16 * commandState,
    _In_ MI_Uint32 exitCode
    );

MI_Uint32 MI_CALL WSManPluginOperationComplete(
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ MI_Uint32 errorCode,
    _In_opt_ const MI_Char16 * extendedInformation
    );

#define WSMAN_PLUGIN_PARAMS_MAX_ENVELOPE_SIZE 1
#define WSMAN_PLUGIN_PARAMS_TIMEOUT 2
#define WSMAN_PLUGIN_PARAMS_REMAINING_RESULT_SIZE 3
#define WSMAN_PLUGIN_PARAMS_LARGEST_RESULT_SIZE 4
#define WSMAN_PLUGIN_PARAMS_GET_REQUESTED_LOCALE 5 /* Returns WSMAN_DATA_TEXT */
#define WSMAN_PLUGIN_PARAMS_GET_REQUESTED_DATA_LOCALE 6 /* Returns WSMAN_DATA_TEXT */


MI_Uint32 MI_CALL WSManPluginGetOperationParameters (
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _Out_ WSMAN_DATA *data
    );

MI_Uint32 MI_CALL WSManPluginGetConfiguration (
  _In_   void * pluginContext,
  _In_   MI_Uint32 flags,
  _Out_  WSMAN_DATA *data
);

MI_Uint32 MI_CALL WSManPluginReportCompletion(
  _In_      void * pluginContext,
  _In_      MI_Uint32 flags
);

MI_Uint32 MI_CALL WSManPluginFreeRequestDetails(_In_ WSMAN_PLUGIN_REQUEST *requestDetails);

#define WSMAN_PLUGIN_STARTUP_REQUEST_RECEIVED     0x0
#define WSMAN_PLUGIN_STARTUP_AUTORESTARTED_REBOOT 0x1
#define WSMAN_PLUGIN_STARTUP_AUTORESTARTED_CRASH  0x2

#define WSMAN_PLUGIN_SHUTDOWN_SYSTEM  1
#define WSMAN_PLUGIN_SHUTDOWN_SERVICE 2
#define WSMAN_PLUGIN_SHUTDOWN_IISHOST 3
#define WSMAN_PLUGIN_SHUTDOWN_IDLETIMEOUT_ELAPSED 4


/* ===================================================================
 * Following functions are plug-in APIs that the shell infrastructure
 * calls into
 * ===================================================================
 */

#define WSMAN_FLAG_SEND_NO_MORE_DATA 1


/* Default signal codes where _EXIT comes from the client to confirm the command has completed and is the last request targeting the command */
#define WSMAN_SIGNAL_CODE_TERMINATE WSMAN_SHELL_NS PAL_T("/signal/Terminate")
#define WSMAN_SIGNAL_CODE_BREAK     WSMAN_SHELL_NS PAL_T("/signal/Break")
#define WSMAN_SIGNAL_CODE_PAUSE     WSMAN_SHELL_NS PAL_T("/signal/Pause")
#define WSMAN_SIGNAL_CODE_RESUME    WSMAN_SHELL_NS PAL_T("/signal/Resume")
#define WSMAN_SIGNAL_CODE_EXIT      WSMAN_SHELL_NS PAL_T("/signal/Exit")


void MI_CALL WSManPluginReleaseShellContext(
    _In_ void * shellContext
    );

void MI_CALL WSManPluginReleaseCommandContext(
    _In_ void * shellContext,
    _In_ void * commandContext
    );

typedef void (MI_CALL *WSManPluginShutdownCallback)(_In_opt_ void *shutdownContext);

void MI_CALL WSManPluginRegisterShutdownCallback(
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ WSManPluginShutdownCallback shutdownCallback,
    _In_opt_ void *shutdownContext);

typedef void (MI_CALL *ShutdownPluginFuncPtr)(_In_ void* pluginContext);

typedef void (MI_CALL *WSManPluginShellFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ MI_Char16 *extraInfo,
    _In_opt_ WSMAN_SHELL_STARTUP_INFO *startupInfo,
    _In_opt_ WSMAN_DATA *inboundShellInformation
    );

typedef void (MI_CALL *WSManPluginConnectFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void* shellContext,
    _In_opt_ void* commandContext,
    _In_opt_ WSMAN_DATA *inboundConnectInformation
    );

typedef void (MI_CALL *WSManPluginReleaseShellContextFuncPtr)(
    _In_ void* pluginContext,
    _In_ void* shellContext
    );

typedef void (MI_CALL *WSManPluginCommandFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void* shellContext,
    _In_ MI_Char16 *commandLine,
    _In_opt_ WSMAN_COMMAND_ARG_SET *arguments
    );

typedef void (MI_CALL *WSManPluginReleaseCommandContextFuncPtr)(
    _In_ void* pluginContext,
    _In_ void* shellContext,
    _In_ void* commandContext
    );

typedef void (MI_CALL *WSManPluginSendFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void* shellContext,
    _In_opt_ void* commandContext,
    _In_ MI_Char16 *stream,
    _In_ WSMAN_DATA *inboundData
    );

typedef void (MI_CALL *WSManPluginReceiveFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void* shellContext,
    _In_opt_ void* commandContext,
    _In_opt_ WSMAN_STREAM_ID_SET* streamSet
    );

typedef void (MI_CALL *WSManPluginSignalFuncPtr)(
    _In_ void* pluginContext,
    _In_ WSMAN_PLUGIN_REQUEST *requestDetails,
    _In_ MI_Uint32 flags,
    _In_ void* shellContext,
    _In_opt_ void* commandContext,
    _In_ MI_Char16 *code);

typedef void (MI_CALL *WSManPluginShellCloseFuncPtr)(
   _In_ void* pluginContext,
   _In_ void* shellContext);

/* List of managed powershell functions we call into to carry out shell operations */
typedef struct _PwrshPluginWkr_Ptrs
{
    ShutdownPluginFuncPtr shutdownPluginFuncPtr;
    WSManPluginShellFuncPtr wsManPluginShellFuncPtr;
    WSManPluginReleaseShellContextFuncPtr wsManPluginReleaseShellContextFuncPtr;
    WSManPluginCommandFuncPtr wsManPluginCommandFuncPtr;
    WSManPluginReleaseCommandContextFuncPtr wsManPluginReleaseCommandContextFuncPtr;
    WSManPluginSendFuncPtr wsManPluginSendFuncPtr;
    WSManPluginReceiveFuncPtr wsManPluginReceiveFuncPtr;
    WSManPluginSignalFuncPtr wsManPluginSignalFuncPtr;
    WSManPluginConnectFuncPtr wsManPluginConnectFuncPtr;
    WSManPluginShellCloseFuncPtr wsManPluginShellCloseFuncPtr;
} PwrshPluginWkr_Ptrs;

/* Function that calls into powershell to get the full set of function pointers */
typedef MI_Uint32 (MI_CALL *InitPluginWkrPtrsFuncPtr)(_Out_ PwrshPluginWkr_Ptrs* wkrPtrs);



#endif /* _SHELL_API_H_ */
